//
//  CurrentDataViewController.swift
//  Chapter 5 HealthKit
//
//  Created by Mini Projects on 16/06/2015.
//  Copyright © 2015 Packt Pub. All rights reserved.
//


import UIKit
import HealthKit

class CurrentDataViewController:UIViewController {
    
    private var weightType:WeightType!
    
    
    private let healthKitStore:HKHealthStore = HKHealthStore()

    private let weightQuantityType = HKQuantityType.quantityTypeForIdentifier(HKQuantityTypeIdentifierBodyMass)!
    private let heartRateQuantityType = HKQuantityType.quantityTypeForIdentifier(HKQuantityTypeIdentifierHeartRate)!
    private let stepCountType = HKQuantityType.quantityTypeForIdentifier(HKQuantityTypeIdentifierStepCount)!
    
    private let sortDescriptor = NSSortDescriptor(key: HKSampleSortIdentifierEndDate,
        ascending: false)
    
    
    private let userDefaults = NSUserDefaults.standardUserDefaults()
    
    @IBOutlet weak var stepCounterLabel: UILabel!
    @IBOutlet var heartRateLabel: UILabel!
    @IBOutlet var weightTextField: UITextField!
    @IBOutlet var weightGoalTextField: UITextField!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        if let weightTypeString = userDefaults.valueForKey("weightType") as? String{
            self.weightType = WeightType(rawValue: weightTypeString)
        } else {
            userDefaults.setValue(WeightType.Kilogram.rawValue, forKey: "weightType")
            self.weightType = .Kilogram
        }
        
        self.readWeightGoal()
        self.readSteps()
        self.readHeartRate()
        self.readWeight()
    }
    
    private func readWeightGoal(){
        M{
            let unitLabel = UILabel()
            if self.weightType == .Kilogram {
                unitLabel.text = "Kg"
            }else if self.weightType == .Stones{
                unitLabel.text = "st."
            }else {
                unitLabel.text = "lb."
            }
            unitLabel.sizeToFit()
            unitLabel.textColor = UIColor.grayColor()
            self.weightGoalTextField.rightView = unitLabel
            self.weightGoalTextField.rightViewMode = .Always

            if let weightGoal = self.userDefaults.valueForKey("weightGoal") {
                self.weightGoalTextField.text = "\(weightGoal)"
            }
        }
    }
    
    
    private func readSteps(){
        let querySteps = HKSampleQuery(sampleType: self.stepCountType, predicate: nil, limit: 1, sortDescriptors: [sortDescriptor]) { query, results, error in
            if let error = error {
                self.displayError(error.localizedDescription)
                return
            }
            if let results = results where results.count > 0{
                let sample = results.first as! HKQuantitySample
                let steps:Int
                steps = Int(sample.quantity.doubleValueForUnit(HKUnit.countUnit()))
                M {
                    self.stepCounterLabel.text = "Number of steps: \(steps)"
                }
            }else{
                M {
                    self.stepCounterLabel.text = "Number of steps: NO DATA"
                }
            }
        }
        healthKitStore.executeQuery(querySteps)
    }
    
    private func readHeartRate(){
        let queryHeartRate = HKSampleQuery(sampleType: self.heartRateQuantityType, predicate: nil, limit: 1, sortDescriptors: [sortDescriptor]) { query, results, error in
            if let error = error {
                self.displayError(error.localizedDescription)
                return
            }
            
            if let results = results where results.count > 0{
                let sample = results.first as! HKQuantitySample
                let bpm:Double
                bpm = sample.quantity.doubleValueForUnit(HKUnit.countUnit().unitDividedByUnit(HKUnit.minuteUnit()))
                M {
                    self.heartRateLabel.text = "Heart Rate: \(bpm)bpm"
                }
            }else{
                M {
                    self.heartRateLabel.text = "Heart Rate: NO DATA"
                }
            }
        }
        healthKitStore.executeQuery(queryHeartRate)
    }
    
    private func readWeight(){
        let queryWeight = HKSampleQuery(sampleType: self.weightQuantityType, predicate: nil, limit: 1, sortDescriptors: [sortDescriptor]) { query, results, error in
            if let error = error {
                self.displayError(error.localizedDescription)
                return
            }
            M{
                let unitLabel = UILabel()
                if self.weightType == .Kilogram {
                    unitLabel.text = "Kg"
                }else if self.weightType == .Stones{
                    unitLabel.text = "st."
                }else {
                    unitLabel.text = "lb."
                }
                unitLabel.sizeToFit()
                unitLabel.textColor = UIColor.grayColor()
                self.weightTextField.rightView = unitLabel
                self.weightTextField.rightViewMode = .Always
            }
            if let results = results where results.count > 0{
                let sample = results.first as! HKQuantitySample
                let weight:Double
                if self.weightType == .Stones {
                    weight = sample.quantity.doubleValueForUnit(HKUnit.stoneUnit())
                }else if self.weightType == .Pounds {
                    weight = sample.quantity.doubleValueForUnit(HKUnit.poundUnit())
                }else {
                    weight = sample.quantity.doubleValueForUnit(HKUnit.gramUnitWithMetricPrefix(.Kilo))
                }
                M {
                    self.weightTextField.text = "\(weight)"
                }
            }else{
                print("No weight")
            }
        }
        healthKitStore.executeQuery(queryWeight)
    }
    
    
    
    private func saveWeightGoal(){
        if let weightGoal = Int(self.weightGoalTextField.text!) {
           userDefaults.setInteger(weightGoal, forKey: "weightGoal")
        }
    }
    
    private func saveWeight(){
        let unit:HKUnit
        
        if self.weightType == .Kilogram{
            unit = HKUnit.gramUnitWithMetricPrefix(.Kilo)
        }else if self.weightType == .Stones {
            unit = HKUnit.stoneUnit()
        } else {
            unit = HKUnit.poundUnit()
        }
        
        if let value =  Double(self.weightTextField.text!) {
            let weightQuantity = HKQuantity(unit: unit,
                doubleValue: value)
            let now = NSDate()
            let sample = HKQuantitySample(type: self.weightQuantityType, quantity: weightQuantity,
                startDate: now, endDate: now)
            healthKitStore.saveObject(sample, withCompletion: { (succeeded, error) in
                if let error = error{
                    print("Failed to save the user's weight: \(error.localizedDescription)")
                }
            })
        }
    }
    
    
    @IBAction func saveAndClose(sender: UIButton) {
        self.saveWeightGoal()
        self.saveWeight()
        self.dismissViewControllerAnimated(true, completion: nil)
    }
    
    @IBAction func changeWeightUnit(sender: UIButton) {
        let alertController = UIAlertController(title: "Weight Unit", message: "Please choose your weight unit", preferredStyle: .ActionSheet)
        let setWeightType = { () in
            self.userDefaults.setValue(self.weightType.rawValue, forKey: "weightType")
            self.readWeight()
        }
        
        let kiloAction = UIAlertAction(title: "Kilograms", style: .Default) { _ in
            self.weightType = .Kilogram
            setWeightType()
        }
        alertController.addAction(kiloAction)
        
        let poundsAction = UIAlertAction(title: "Pounds", style: .Default) { _ in
            self.weightType = .Pounds
            setWeightType()
        }
        alertController.addAction(poundsAction)
        
        let stonesAction = UIAlertAction(title: "Stones", style: .Default) { _ in
            self.weightType = .Stones
            setWeightType()
        }
        alertController.addAction(stonesAction)
        
        presentViewController(alertController, animated: true, completion: nil)
    }
} // end CurrentDataViewController
