//
//  WeeklyChartsViewController.swift
//  Chapter 5 HealthKit
//
//  Created by Mini Projects on 19/06/2015.
//  Copyright © 2015 Packt Pub. All rights reserved.
//

import UIKit
import HealthKit
import Charts

class WeeklyChartsViewController: UIViewController{
    
    @IBOutlet var stepsChart: LineChartView!
    @IBOutlet var heartBeatChart: LineChartView!
    @IBOutlet var weightChart: LineChartView!
    
    
    private let healthKitStore:HKHealthStore = HKHealthStore()
    private let stepCountType = HKQuantityType.quantityTypeForIdentifier(HKQuantityTypeIdentifierStepCount)!
    private let heartRateQuantityType = HKQuantityType.quantityTypeForIdentifier(HKQuantityTypeIdentifierHeartRate)!
    private let weightQuantityType = HKQuantityType.quantityTypeForIdentifier(HKQuantityTypeIdentifierBodyMass)!
    
    
    private let sortDescriptor = NSSortDescriptor(key: HKSampleSortIdentifierEndDate,
        ascending: true)
    private var predicate:NSPredicate!
    
    
    private var userDefaults = NSUserDefaults.standardUserDefaults()
    private var weightType:WeightType!
    
    override func viewDidLoad() {
        let currentDate = WeeklyChartsViewController.stripDate(NSDate())
        self.predicate = NSPredicate(format: "endDate >= %@", currentDate.dateByAddingTimeInterval(-6 * 24 * 60 * 60))
        self.readSteps()
        self.readHeartRate()
        self.readWeight()
    }
    
    private func readSteps(){
        let querySteps = HKSampleQuery(sampleType: self.stepCountType, predicate: predicate, limit: 0, sortDescriptors: [sortDescriptor]) { query, results, error in
            if let error = error {
                self.displayError(error.localizedDescription)
                return
            }
            if let results = results where results.count > 0{
                var stepSum = [String:Double]()
                let dateFormatter = NSDateFormatter()
                dateFormatter.dateFormat = "MM/dd"
                for sample in results as! [HKQuantitySample]{
                    let steps = sample.quantity.doubleValueForUnit(HKUnit.countUnit())

                    
                    let dateString = dateFormatter.stringFromDate(sample.endDate)
                    if let _ = stepSum[dateString] {
                        stepSum[dateString]! += steps
                    }else {
                        stepSum[dateString] = steps
                    }
                }
                M{
                    self.stepsChart.descriptionText = "Steps"
                    self.stepsChart.noDataTextDescription = "There is no step data"
                    self.stepsChart.highlightEnabled = false
                    self.stepsChart.dragEnabled = false
                    self.stepsChart.setScaleEnabled(false)
                    self.stepsChart.pinchZoomEnabled = false
                    self.stepsChart.drawGridBackgroundEnabled = false
                    
                    
                    let xVals = stepSum.keys.sort()
                    var yVals = [ChartDataEntry]()
                    for (i, key) in xVals.enumerate() {
                        yVals.append(ChartDataEntry(value: stepSum[key]!, xIndex: i))
                    }
                    
                    let dataset = LineChartDataSet(yVals: yVals, label: "Number of Steps")
                    self.stepsChart.data = LineChartData(xVals: xVals, dataSets: [dataset])
                }
            }
        }
        healthKitStore.executeQuery(querySteps)
    }
    
    
    private func readHeartRate(){
        let queryHeartRate = HKSampleQuery(sampleType: self.heartRateQuantityType, predicate: predicate, limit: 0, sortDescriptors: [sortDescriptor]) { query, results, error in
            if let error = error {
                self.displayError(error.localizedDescription)
                return
            }
            if let results = results where results.count > 0{
                var xVals = [String]()
                var yVals = [ChartDataEntry]()
                let dateFormatter = NSDateFormatter()
                dateFormatter.dateFormat = "MM/dd"
                for sample in results as! [HKQuantitySample]{
                    let heartRate = sample.quantity.doubleValueForUnit(HKUnit.countUnit().unitDividedByUnit(HKUnit.minuteUnit()))
                    let dateString = dateFormatter.stringFromDate(sample.endDate)
                    xVals.append(dateString)
                    yVals.append(ChartDataEntry(value: heartRate, xIndex: yVals.count))
                }
                M{
                    self.heartBeatChart.descriptionText = "Heart Rate"
                    self.heartBeatChart.noDataTextDescription = "There is no heart rate data"
                    self.heartBeatChart.highlightEnabled = false
                    self.heartBeatChart.dragEnabled = false
                    self.heartBeatChart.setScaleEnabled(false)
                    self.heartBeatChart.pinchZoomEnabled = false
                    self.heartBeatChart.drawGridBackgroundEnabled = false

                    
                    let dataset = LineChartDataSet(yVals: yVals, label: "Heart Rate")
                    
                    self.heartBeatChart.data = LineChartData(xVals: xVals, dataSets: [dataset])
                }
            }
        }
        healthKitStore.executeQuery(queryHeartRate)
    }
    
    private func readWeightUnit(){
        if let weightTypeString = userDefaults.valueForKey("weightType") as? String{
            self.weightType = WeightType(rawValue: weightTypeString)
        } else {
            userDefaults.setValue(WeightType.Kilogram.rawValue, forKey: "weightType")
            self.weightType = .Kilogram
        }
    }
    
    private func readWeightGoal() -> Double?{
        return self.userDefaults.valueForKey("weightGoal") as? Double
    }
    
    
    private func readWeight(){
        self.readWeightUnit()
        let querySteps = HKSampleQuery(sampleType: self.weightQuantityType, predicate: predicate, limit: 0, sortDescriptors: [sortDescriptor]) { query, results, error in
            if let error = error {
                self.displayError(error.localizedDescription)
                return
            }
            if let results = results where results.count > 0{
                var firstWeight:Double?
                var lastWeight:Double?
                var weightTable = [String:[Double] ]()
                let dateFormatter = NSDateFormatter()
                dateFormatter.dateFormat = "MM/dd"
                for sample in results as! [HKQuantitySample]{
                    let weight:Double
                    switch self.weightType.rawValue {
                    case WeightType.Stones.rawValue:
                        weight = sample.quantity.doubleValueForUnit(HKUnit.stoneUnit())
                    case WeightType.Pounds.rawValue:
                        weight = sample.quantity.doubleValueForUnit(HKUnit.poundUnit())
                    default:
                        weight = sample.quantity.doubleValueForUnit(HKUnit.gramUnitWithMetricPrefix(.Kilo))
                    }
                    if firstWeight == nil {
                        firstWeight = weight
                    }
                    lastWeight = weight


                    
                    let dateString = dateFormatter.stringFromDate(sample.endDate)
                    if let _ = weightTable[dateString] {
                        weightTable[dateString]!.append(weight)
                    }else {
                        weightTable[dateString] = [weight]
                    }
                }
                M{
                    self.weightChart.descriptionText = "Weight"
                    self.weightChart.noDataTextDescription = "There is no weight data"
                    self.weightChart.highlightEnabled = false
                    self.weightChart.dragEnabled = false
                    self.weightChart.setScaleEnabled(false)
                    self.weightChart.pinchZoomEnabled = false
                    self.weightChart.drawGridBackgroundEnabled = false
                    
                    
                    let xVals = weightTable.keys.sort()
                    var yVals = [ChartDataEntry]()
                    for (i, key) in xVals.enumerate() {
                        let weightArray = weightTable[key]!
                        let weightSum = weightArray.reduce(0, combine: { (sum, element) -> Double in
                            return sum + element
                        })
                        
                        let avg = weightSum / Double(weightArray.count)
                        yVals.append(ChartDataEntry(value: avg , xIndex: i))
                    }
                    
                    let dataset = LineChartDataSet(yVals: yVals, label: "Weight")
                    
                    self.weightChart.data = LineChartData(xVals: xVals, dataSets: [dataset])
                    
                    if results.count > 1 {
                        if let goal = self.readWeightGoal() {
                            let lastDiff = abs(goal - lastWeight!)
                            let firstDiff = abs(goal - firstWeight!)
                            if lastDiff == 0 {
                                self.displayMessage("Congrats, you have your desired weight")
                            }else if lastDiff < firstDiff {
                                self.displayMessage("You are doing well")
                            }else {
                                self.displayMessage("Be carefull! You are moving away from your desired weight")
                            }
                        }else{
                            self.displayMessage("No desired weight is set")
                        }
                    }
                }
            }
        }
        healthKitStore.executeQuery(querySteps)
    }
    
    // MARK: - Auxiliary functions
    
    private func displayMessage(message:String){
        let alertController = UIAlertController(title: "Message", message: message, preferredStyle: .Alert)
        let alertAction = UIAlertAction (title: "Dismiss", style: .Cancel, handler: nil)
        alertController.addAction(alertAction)
        self.presentViewController(alertController, animated: true, completion: nil)
    }
    
    private class func stripDate(date: NSDate) -> NSDate {
        let currentCalendar = NSCalendar.currentCalendar()
        let components = currentCalendar.components(NSCalendarUnit.Year.union(.Month).union(.Day), fromDate: date)
        return currentCalendar.dateFromComponents(components)!
    }
    
    // MARK: UI methods
    @IBAction func close(sender: AnyObject) {
        self.dismissViewControllerAnimated(true, completion: nil)
    }

} // end WeeklyChartsViewController