//
//  RecorderViewController.swift
//  Chapter 8 Rehash
//
//  Created by City Life on 16/08/2015.
//  Copyright © 2015 PacktPub. All rights reserved.
//

import Foundation
import UIKit
import AVFoundation
import Photos

class RecorderViewController: UIViewController, AVAudioRecorderDelegate {
    
    @IBOutlet weak var videoView: UIView!
    @IBOutlet weak var recordButton: UIButton!
    
    var assetInfo:AssetInfo!
    var recorder:AVAudioRecorder!
    var player: AVPlayer!
    var playerItem:AVPlayerItem!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        self.initRecorder()
    }
    
    private func initRecorder() {
        if let audioUrl = try! self.assetInfo.prepareAudio() {
            let session = AVAudioSession.sharedInstance()
            do {
                try session.setCategory(AVAudioSessionCategoryPlayAndRecord)
                try session.setActive(true)
                
                var recordSetting = [String:AnyObject]()
                recordSetting[AVFormatIDKey] =  NSNumber(unsignedInt: kAudioFormatMPEG4AAC)
                recordSetting[AVSampleRateKey] = 44100.0
                recordSetting[AVNumberOfChannelsKey] = 2
                
                recorder = try AVAudioRecorder(URL: audioUrl, settings: recordSetting)
                recorder.delegate = self
                recorder.meteringEnabled = true
                recorder.prepareToRecord()
                self.recordButton.enabled = true
            }catch  {
                print("Fail to initialize the recorder. File: \(__FILE__), Line: \(__LINE__)")
            }
        }
    } // end initRecord
    
    @IBAction func startRecording(sender: UIButton) {
        // avoid letting the user press twice the same button.
        sender.enabled = false
        
        let asset = AVAsset(URL: self.assetInfo.videoUrl)
        playerItem = AVPlayerItem(asset: asset)
        playerItem.addObserver(self, forKeyPath: "status", options: .New, context: nil)
        NSNotificationCenter.defaultCenter().addObserver(self, selector: Selector("itemDidFinishPlaying:"), name: AVPlayerItemDidPlayToEndTimeNotification, object: playerItem)

        player = AVPlayer(playerItem: playerItem)
        player.actionAtItemEnd = AVPlayerActionAtItemEnd.Pause
        
        let playerLayer = AVPlayerLayer(player: player)
        playerLayer.frame = self.videoView.bounds
        self.videoView.layer.addSublayer(playerLayer)
    }
    
    
    override func observeValueForKeyPath(keyPath: String?, ofObject object: AnyObject?, change: [String : AnyObject]?, context: UnsafeMutablePointer<Void>) {
        if let playerItem = object as? AVPlayerItem where playerItem.status == .ReadyToPlay {
            player.play()
            recorder.record()
        }else {
            super.observeValueForKeyPath(keyPath, ofObject: object, change: change, context: context)
        }
    }
    
    func itemDidFinishPlaying(notification: NSNotification){
        recorder.stop()
    }
    
    
    func audioRecorderDidFinishRecording(recorder: AVAudioRecorder, successfully flag: Bool) {
        if flag {
            self.merge()
        }else {
            print("Record failed")
        }
    }
    
    func merge() {
        do {
            let composition = AVMutableComposition()
            
            // Audio part
            let audioAsset = AVURLAsset(URL: self.assetInfo.audioUrl!)
            let audioTimeRange = CMTimeRangeMake(kCMTimeZero, audioAsset.duration)
            let compositionAudioTrack = composition.addMutableTrackWithMediaType(AVMediaTypeAudio, preferredTrackID: kCMPersistentTrackID_Invalid)
            try compositionAudioTrack.insertTimeRange(audioTimeRange, ofTrack: audioAsset.tracksWithMediaType(AVMediaTypeAudio).first!, atTime: kCMTimeZero)

            //Video part
            let videoAsset = AVURLAsset(URL: assetInfo.videoUrl)
            let videoTimeRange = CMTimeRangeMake(kCMTimeZero,videoAsset.duration)
            let compositionVideoTrack = composition.addMutableTrackWithMediaType(AVMediaTypeVideo, preferredTrackID: kCMPersistentTrackID_Invalid)
            let videoTrack = videoAsset.tracksWithMediaType(AVMediaTypeVideo).first!
            compositionVideoTrack.preferredTransform = videoTrack.preferredTransform
            try compositionVideoTrack.insertTimeRange(videoTimeRange, ofTrack: videoTrack, atTime: kCMTimeZero)
            
            
            
            // Video Output
            let fileManager = NSFileManager.defaultManager()
            let globallyUniqueString = NSProcessInfo.processInfo().globallyUniqueString
            let tempPath = (NSTemporaryDirectory() as NSString).stringByAppendingPathComponent(globallyUniqueString)
            try fileManager.createDirectoryAtPath(tempPath, withIntermediateDirectories: false, attributes: nil)
            let outputPath = (tempPath as NSString).stringByAppendingPathComponent("rehash-output.mov")
            let outputFileUrl = NSURL(fileURLWithPath: outputPath)

            // Fusion
            let assetExport = AVAssetExportSession(asset: composition, presetName: AVAssetExportPresetHighestQuality)
            assetExport?.outputFileType = "com.apple.quicktime-movie"
            assetExport?.outputURL = outputFileUrl
            assetExport?.exportAsynchronouslyWithCompletionHandler({ () -> Void in
                PHPhotoLibrary.sharedPhotoLibrary().performChanges({ () -> Void in
                    PHAssetChangeRequest.creationRequestForAssetFromVideoAtFileURL(outputFileUrl)
                    }, completionHandler: { (success:Bool, error:NSError?) -> Void in
                        if success {
                            try! fileManager.removeItemAtPath(outputPath)
                            print("Successful saved to photo library")
                        }else {
                            print("Couldn't save to photo library")
                        }
                        dispatch_async(dispatch_get_main_queue(), { () -> Void in
                            self.performSegueWithIdentifier("exitrecorder", sender: self)
                        })
                })
            })
        } catch {
            print("Failed merging the audio and video")
        }
    } // end merge
    
    deinit {
        self.playerItem.removeObserver(self, forKeyPath: "status")
        NSNotificationCenter.defaultCenter().removeObserver(self.playerItem)
    }
    
    
} // end RecorderViewController
